#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "she.h"
#include "reflection.h"
#include "hashtab.h"
#include "objalloc.h"

extern char* yytext;
extern FILE* yyin;
int yylex();

typedef reflection_type_t she_type;

typedef struct {
    reflection_value_t value;
    she_type type;
    char* name;
} she_value;

static int hash_string_eq(const void *s1_p, const void *s2_p) {
    const she_value* s1 = (const she_value*)s1_p;
    const char* s2 = (const char*)s2_p;
    return strcmp(s1->name, s2) == 0;
}

reflection_t* ref;
char* cmd;
int running;
she_value args[4];
int arg_num;
htab_t values;
char* ret_val;
struct objalloc* allocator;

reflection_value_t wait_ident();

void err(char* msg, ...) {
    va_list ap;
    va_start(ap, msg);
    fprintf(stderr, "error: ");
    vfprintf(stderr, msg, ap);
    fprintf(stderr, "\n");
    exit(1);
}

static char* mstrcpy(const char* s) {
    char* ret = (char*)objalloc_alloc(allocator, sizeof(char)*(strlen(s)+1));
    strcpy(ret, s);
    return ret;
}

she_type get_ret_type() {
    if (!ret_val) return R_VOID;

    switch (ret_val[0]) {
    case 'c':
        return R_CHAR;
    case 'i':
        return R_INT;
    case 'f':
        return R_FLOAT;
    case 'p':
        return R_PTR;
#ifdef HAVE_FFI
    case 'd':
        return R_DOUBLE;
#endif
    default:
        err("invalid valuable name");
        return R_VOID;  /* dummy */
    }
}

void dump_value(she_value val) {
    switch (val.type) {
    case R_INT:
        printf("%d\n", val.value.i); break;
    case R_FLOAT:
        printf("%f\n", val.value.f); break;
#ifdef HAVE_FFI
    case R_DOUBLE:
        printf("%f\n", val.value.d); break;
#endif
    case R_PTR:
        printf("%p\n", val.value.p); break;
    case R_CHAR:
        printf("%c\n", val.value.c); break;
    default:
        err("unknown type");
    }
}

int get_tok() {
    int t = yylex();
    if (t == 0) running = 0;
    return t;
}

int wait_args() {
    int tok;

    while (1) {
        tok = get_tok();

        switch ((int)tok) {
        case '=': {
            reflection_value_t ret;
            she_value* val;
            she_value** valp;

            ret_val = cmd;
            ret = wait_ident();
            val = (she_value*)malloc(sizeof(she_value));
            val->value = ret;
            val->type = get_ret_type();
            val->name = ret_val;

            valp = (she_value**)htab_find_slot(values, val->name, INSERT);
            *valp = val;

            ret_val = NULL;

            return 1;
        }
        case '\n':
        case 0:
            return 0;
        case TOK_INT:
            args[arg_num].type = R_INT;
            args[arg_num++].value.i = atoi(yytext);
            break;
        case TOK_FLOAT:
            args[arg_num].type = R_DOUBLE;
            args[arg_num++].value.d = atof(yytext);
            break;
        case TOK_STRING:
            args[arg_num].type = R_PTR;
            yytext[strlen(yytext)-1] = '\0';
            args[arg_num++].value.p = mstrcpy(yytext+1);
            break;
        case TOK_IDENT: {
            she_value** valp;
            valp = (she_value**)htab_find_slot(values, yytext, NO_INSERT);
            if (!valp) {
                err("unknown valuable");
            }
            args[arg_num++] = **valp;
            break;
        }
        }
    }
}

void wait_sys_cmd() {
    int tok = get_tok();

    if (tok != TOK_IDENT) {
        err("not system command");
    }

    cmd = mstrcpy(yytext);
    if (wait_args()) {
        err("mysterious = token");
    }

    if (strcmp(cmd, "load") == 0) {
        if (arg_num != 1 && arg_num != 2) {
            err("!load needs 1 or 2 argument");
        }
        if (args[0].type != R_PTR) {
            err("!load needs string argument");
        }
        if (arg_num == 2) {
            if (args[1].type != R_PTR) {
                err("!load needs string argument");
            }
        }
        else {
            args[1].value.p = NULL;
        }
        if (reflection_dlopen(ref, args[0].value.p, args[1].value.p)) {
            err("%s: load fail", args[0].value.p);
        }
    }
    else if (strcmp(cmd, "print") == 0) {
        if (arg_num != 1) {
            err("!print needs 1 argument");
        }
        dump_value(args[0]);
    }
    else if (strcmp(cmd, "dump_symbols") == 0) {
        if (arg_num != 0) {
            err("!dump_symbols needs no argument");
        }
        reflection_dump_symbols(ref);
    }
    else if (strcmp(cmd, "exit") == 0) {
        running = 0;
    }
    else {
        err("%s: unknown system command", cmd);
    }
    cmd = NULL;
}

reflection_value_t wait_ident() {
    int tok;

    reflection_value_t ret;

    while ((tok = get_tok()) == '\n');

    if (tok == TOK_IDENT) {
        cmd = mstrcpy(yytext);

        if (!wait_args()) {
            reflection_value_t values[4];
            reflection_type_t types[4];
            reflection_type_t ret_type = get_ret_type();
            int i;

            for (i = 0; i < arg_num; i++) {
                values[i] = args[i].value;
                types[i] = args[i].type;
            }
            /* printf("cmd=%s argnum=%d\n", cmd, arg_num); */
            if (reflection_dispatch(ref, cmd, ret_type,
                                    arg_num, values, types, &ret))
            {
                err("%s: fail", cmd);
            }
        }
        cmd = NULL;
    }
    else if (tok == '!') {
        wait_sys_cmd();
    }
    else if (tok != 0) {
        err("not ident");
    }

    return ret;
}

int main(int argc, char* argv[]) {
    allocator = objalloc_create();

    ref = reflection_init(argv[0], 20001);
    if (!ref) {
//        err("cannot init reflection");
    }

    if (argc >= 2) {
        yyin = fopen(argv[1], "rb");
        if (yyin == NULL) {
            err("%s: cannot open", argv[1]);
        }
    }

    values = htab_create(101, htab_hash_string, hash_string_eq, NULL);

    running = 1;
    while (running) {
        arg_num = 0;
        ret_val = NULL;
        wait_ident();
    }

    reflection_quit(ref);
    objalloc_free(allocator);

    return 0;
}
